/* -*-c++-*- source - Open Producer - Copyright (C) 2002 Don Burns
 * Distributed under the terms of the GNU LIBRARY GENERAL PUBLIC LICENSE (LGPL)
 * as published by the Free Software Foundation.
 */


#ifdef _X11_IMPLEMENTATION
#  include <X11/Xlib.h>
#endif

#include <math.h>

#include <Producer/Math>
#include <Producer/CameraConfig>

using namespace Producer;

unsigned int CameraConfig::getNumberOfScreens()
{
#ifdef _X11_IMPLEMENTATION
    Display *dpy = XOpenDisplay("");
    if( dpy == NULL )
    {
        std::cerr << "Unable to open display \"" << XDisplayName( ":0.0") << "\".  Is the DISPLAY environmental variable set?\n";
        return 0;
    }
    unsigned int nscreens = ScreenCount(dpy); 
    XCloseDisplay(dpy);
    return nscreens;
#else
    return 1;
#endif
}



void CameraConfig::rotateCameraOffset( float deg, float x, float y, float z )
{
#define deg2rad(x) ((x)*M_PI/180)
    Matrix m;
    m.invert(Matrix::rotate( deg2rad(deg), x,y,z)); 
    m = m * Matrix(_offset_matrix);
    memcpy( _offset_matrix, m.ptr(), sizeof( float[16] ));
}

void CameraConfig::translateCameraOffset( float x, float y, float z )
{
    Matrix m;
    m.invert(Matrix::translate( x,y,z)); 
    m = m * Matrix(_offset_matrix);
    memcpy( _offset_matrix, m.ptr(), sizeof( float[16] ));
}

void CameraConfig::scaleCameraOffset( float x, float y, float z )
{
    Matrix m = Matrix::scale( x,y,z) * Matrix(_offset_matrix);
    memcpy( _offset_matrix, m.ptr(), sizeof( float[16] ));
}

#if 0
#ifdef WIN32
bool CameraConfig::parseFile(  const char *file )
{
    std::cerr << "Sorry... Config file parsing not available for Win32 yet.\n";
    return false;
}
#endif
#endif
