/* -*-c++-*- source - Open Producer - Copyright (C) 2002 Don Burns
 * Distributed under the terms of the GNU LIBRARY GENERAL PUBLIC LICENSE (LGPL)
 * as published by the Free Software Foundation.
 */
#include <Producer/RenderSurface>

using namespace std;
using namespace Producer;


RenderSurface::RenderSurface( void )
{
    _hostname        = "";
    _displayNum     = 0;
    _screen          = 0;
    _mayFullScreen   = true;
    _isFullScreen    = true;
#ifdef _X11_IMPLEMENTATION
    char *envptr = getenv( "DISPLAY" );
    if( envptr != NULL && *envptr != 0 )
    {
        size_t p0 = 0;
        size_t p1 = string(envptr).find(":", p0);
        _hostname = string(envptr).substr(p0,p1);
        p0 = p1+1;
        p1 = string(envptr).find(".", p0);

        if( p1 > 0 )
        {
            _displayNum = atoi((string(envptr).substr(p0,p1)).c_str());
            p0 = p1+1;
            p1 = string(envptr).length() - p0;
            if( p1 > 0 )
                _screen = atoi((string(envptr).substr(p0,p1)).c_str());
        }
        else if( p1 < string(envptr).length() )
        {
            p1 = string(envptr).length();
            _displayNum = atoi((string(envptr).substr(p0,p1)).c_str());
            _screen = 0;
        }
    }
#endif
    _windowLeft     = 0;
    _windowRight    = 1;
    _windowBottom   = 0;
    _windowTop      = 1; 
    _windowX        = 0;
    _windowY        = 0;
    _windowWidth    = 0xFFFFFFFF;
    _windowHeight   = 0xFFFFFFFF;
    _dpy             = NULL;
    _win             = 0;
    _parent          = 0;
    _visualChooser  = NULL;
    _visualInfo     = NULL;
    _decorations     = true;
    _useCursor       = true;
    _windowName     = " *** RenderSurface *** ";
    _realized        = false;
    _useConfigEventThread = true;
    _threadReady    = NULL;
    _useDefaultEsc = true;
    
#ifdef _WIN32_IMPLEMENTATION
    _ownWindow = true;
    _ownVisualChooser = true;
    _ownVisualInfo = true;
    _hinstance = NULL;
    _glcontext = NULL;
    _hdc = NULL;
    _mx = 0;
    _my = 0;
    _mbutton = 0;
#endif
}

RenderSurface::~RenderSurface( void )
{
#ifdef _X11_IMPLEMENTATION
    if( _realized )
        XDestroyWindow(_dpy, _win );
    if( _visualInfo != NULL )
        XFree(_visualInfo);
#endif
#ifdef _WIN32_IMPLEMENTATION
#endif
}

void RenderSurface::setParentWindow( Window parent )
{
     _parent = parent;
}

Window RenderSurface::getParentWindow( void ) const
{
    return _parent;
}

void RenderSurface::setWindowName( const std::string &name )
{
    _windowName = name;
}

const std::string &RenderSurface::getWindowName( void ) const
{
    return _windowName;
}

void RenderSurface::setHostName( const std::string &name )
{
    _hostname = name;
}

const std::string &RenderSurface::getHostName( void ) const
{
    return _hostname;
}

void RenderSurface::setDisplay( Display *dpy )
{
    _dpy = dpy;
}

Display *RenderSurface::getDisplay()
{
    return _dpy;
}

const Display *RenderSurface::getDisplay() const
{
    return _dpy;
}

void RenderSurface::setDisplayNum( int num )
{
    _displayNum = num;
}
int  RenderSurface::getDisplayNum( void ) const
{
    return _displayNum;
}

void RenderSurface::setScreenNum( int num )
{
    _screen = num;
}

int  RenderSurface::getScreenNum( void ) const
{
    return _screen;
}

#if 0
void RenderSurface::setWindowRect( float left, float right, 
                                   float bottom, float top) 
{
    _window_left   = left;
    _window_right  = right;
    _window_bottom = bottom;
    _window_top    = top;
    _isFullScreen    = false;
    if( _realized )
        _resizeWindow();
}

void RenderSurface::getWindowRect( float &left, float &right,
                                   float &bottom, float &top) const
{
    if( _isFullScreen )
    {
        left = 0.0f;
        right = 1.0f;
        bottom = 0.0f;
        top = 1.0f;
    }
    else
    {
        left   = _window_left;
        right  = _window_right;
        bottom = _window_bottom;
        top    = _window_top;
    }
}
#endif

void RenderSurface::setWindowRectangle( int x, int y, unsigned int width, unsigned int height )
{
    _windowX = x;
    _windowY = y;
    _windowWidth = width;
    _windowHeight = height;
    _isFullScreen    = false;
    if( _realized )
        _resizeWindow();
}

void RenderSurface::getWindowRectangle( int &x, int &y, unsigned int &width, unsigned int &height ) const
{
    if( _isFullScreen )
    {
        x = 0;
        y = 0;
        width = _screenWidth;
        height = _screenHeight;
    }
    else
    {
        x      = _windowX;
        y      = _windowY;
        width  = _windowWidth;
        height = _windowHeight;
    }
}

int RenderSurface::getWindowOriginX()
{
    if( _isFullScreen ) return 0;
    return _windowX;
}

int RenderSurface::getWindowOriginY()
{
    if( _isFullScreen ) return 0;
    return _windowY;
}

unsigned int RenderSurface::getWindowWidth() const
{
    if( _isFullScreen ) return _screenWidth;
    return _windowWidth;
}

unsigned int RenderSurface::getWindowHeight() const
{
    if( _isFullScreen ) return _screenHeight;
    return _windowHeight;
}

void RenderSurface::setInputRectangle( const InputRectangle &inputRectangle )
{
    _inputRectangle = inputRectangle;
}

const RenderSurface::InputRectangle &RenderSurface::getInputRectangle()
{
    return _inputRectangle;
}


void RenderSurface::useBorder( bool flag )
{
    _decorations = flag;
    if( _realized )
          _setBorder(_decorations);
}

bool RenderSurface::usesBorder()
{
    return _decorations;
}

void RenderSurface::useCursor( bool flag )
{
    _useCursor = flag;
    if( _realized )
        _setCursor();
}

void RenderSurface::setWindow( const Window win )
{
    if( _realized )
    {
        std::cerr << "RenderSurface::setWindow() - cannot set window after RenderSurface has been realized\n";
        return;
    }
    _win = win;
#ifdef WIN32
    _ownWindow = false;
#endif
}

Window RenderSurface::getWindow( void ) const
{
    return _win;
}


GLContext RenderSurface::getGLContext( void) const
{
    return _glcontext;
}


bool RenderSurface::isRealized() const
{
    return _realized;
}

void RenderSurface::setVisualChooser( VisualChooser *vc )
{
    if( _realized )
    {
        std::cerr << "RenderSurface::setVisualChooser():Warning - has no effect after RenderSurface has been realized\n";
        return;
    }
    _visualChooser = vc;
#ifdef WIN32
    _ownVisualChooser = false;
#endif
}

VisualChooser *RenderSurface::getVisualChooser( void )
{
    return _visualChooser;
}

const VisualChooser *RenderSurface::getVisualChooser( void ) const
{
    return _visualChooser;
}

void RenderSurface::setVisualInfo( VisualInfo *vi )
{
    if( _realized )
    {
        std::cerr << "RenderSurface::setVisualInfo():Warning - has no effect after RenderSurface has been realized\n";
        return;
    }
    _visualInfo = vi;
#ifdef WIN32
    _ownVisualInfo = false;
#endif
}

VisualInfo *RenderSurface::getVisualInfo( void  )
{
    return _visualInfo;
}

const VisualInfo *RenderSurface::getVisualInfo( void  ) const
{
    return _visualInfo;
}

void  RenderSurface::useConfigEventThread( bool flag )
{
    if( _realized )
    {
        // This message is annoying.  Cameras that share a render surface will call this after the
        // render surface has been realized, which is valid.
        //std::cerr << "RenderSurface::useConfigEventThread():Warning - has no effect after RenderSurface has been realized\n";
        return;
    }
    _useConfigEventThread = flag;
}

unsigned int RenderSurface::getRefreshRate() const
{
    if( !_realized ) return 0;
    return _refreshRate; 
}

void RenderSurface::setRealizeCallback( Callback *realizeCB )
{
    if( _realized ) 
    {
        std::cerr << "RenderSurface::setRealizeCallback() : Warning.  RenderSurface is already realized.  Setting ignored.\n";
            return;
    }
    _realizeCB = realizeCB;
}

bool RenderSurface::waitForRealize()
{
    if( _realized ) return true;
    while( _realized == false )
        _realizeBlock.block();
    return true;
}

void RenderSurface::positionPointer( int x, int y )
{
    _positionPointer(x,y);
}
