/* -*-c++-*- source - Open Producer - Copyright (C) 2002 Don Burns
 * Distributed under the terms of the GNU LIBRARY GENERAL PUBLIC LICENSE (LGPL)
 * as published by the Free Software Foundation.
 */


#ifdef WIN32 

#include <iostream>
#include <map>
#include <windows.h>
#include <Producer/RenderSurface>
#include <Producer/Events>
#include <Producer/Keyboard>

// Use the GLUT CURSOR defines for now
#define  GLUT_CURSOR_RIGHT_ARROW            0x0000
#define  GLUT_CURSOR_LEFT_ARROW             0x0001
#define  GLUT_CURSOR_INFO                   0x0002
#define  GLUT_CURSOR_DESTROY                0x0003
#define  GLUT_CURSOR_HELP                   0x0004
#define  GLUT_CURSOR_CYCLE                  0x0005
#define  GLUT_CURSOR_SPRAY                  0x0006
#define  GLUT_CURSOR_WAIT                   0x0007
#define  GLUT_CURSOR_TEXT                   0x0008
#define  GLUT_CURSOR_CROSSHAIR              0x0009
#define  GLUT_CURSOR_UP_DOWN                0x000A
#define  GLUT_CURSOR_LEFT_RIGHT             0x000B
#define  GLUT_CURSOR_TOP_SIDE               0x000C
#define  GLUT_CURSOR_BOTTOM_SIDE            0x000D
#define  GLUT_CURSOR_LEFT_SIDE              0x000E
#define  GLUT_CURSOR_RIGHT_SIDE             0x000F
#define  GLUT_CURSOR_TOP_LEFT_CORNER        0x0010
#define  GLUT_CURSOR_TOP_RIGHT_CORNER       0x0011
#define  GLUT_CURSOR_BOTTOM_RIGHT_CORNER    0x0012
#define  GLUT_CURSOR_BOTTOM_LEFT_CORNER     0x0013
#define  GLUT_CURSOR_INHERIT                0x0064
#define  GLUT_CURSOR_NONE                   0x0065
#define  GLUT_CURSOR_FULL_CROSSHAIR         0x0066

    // Set the cursor AND change it for this window class. 
#define MAP_CURSOR(a,b) case a: SetCursor( LoadCursor(NULL,b) ); \
SetClassLong(_win,GCL_HCURSOR,LONG(LoadCursor(NULL,b))); \
break;
    // Nuke the cursor AND change it for this window class. 
#define ZAP_CURSOR(a,b) case a: SetCursor(NULL); \
SetClassLong(_win,GCL_HCURSOR,LONG(NULL)); \
break;


using namespace Producer;

// Class statics
std::map <Window, RenderSurface *>RenderSurface::registry;


class RenderSurface::Client 
{
    public:
        Client(unsigned long mask): _mask(mask) {}
        unsigned int mask() { return _mask; }
        EventQueue *getQueue() { return &q; }
        void queue( ref_ptr<Event> ev ) { q.push_back( ev ); }

    private :
        unsigned int _mask;
        EventQueue q;
};

LONG WINAPI RenderSurface::s_proc( Window hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
    std::map <Window, RenderSurface *>::iterator p;
    p = registry.find( hWnd );
    if( p == registry.end() )
        return DefWindowProc( hWnd, uMsg, wParam, lParam );
    else
        return p->second->proc( hWnd, uMsg, wParam, lParam );
	
}

void RenderSurface::dispatch( ref_ptr<Event> ev )
{
    if( clients.size() == 0 ) return;

    std::vector <Client *>::iterator p;
    for( p = clients.begin(); p != clients.end(); p++ )
    {
#if 0
        if( ev->type() == Event::MouseMotion )
        {
            // If MouseMotionMask is set, then send event regardless of
            // Mouse button state.
            if( (*p)->mask() & Event::MouseMotionMask )
                (*p)->queue(ev);
            // Otherwise, send event only if mouse button is pressed
            else if( (*p)->mask() & Event::ButtonMotionMask && _mbutton )
                (*p)->queue(ev);
        }
        else if( (*p)->mask() & (1<<ev->type()) )
            (*p)->queue( ev ); 
#endif

        if( (*p)->mask() & (1<<ev->type()) )
		{		
            (*p)->queue( ev ); 
		}
    } 
}


LONG WINAPI RenderSurface::proc( Window hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
    LONG lRet = 1;
    RECT rect;
    ref_ptr<Event> ev;

    switch (uMsg)
    {
//STUPID $%#&*% Windows
#ifndef WM_MOUSEWHEEL
#define WM_MOUSEWHEEL 522
#endif
	case WM_MOUSEWHEEL:
		{
			KeyboardMouseCallback::ScrollingMotion scroll = 
				(signed int)(wParam) < 0 ? KeyboardMouseCallback::ScrollDown:
				KeyboardMouseCallback::ScrollUp;
			ev = new MouseScrollEvent(hWnd,scroll);
			dispatch(ev);
		}
		break;
    case WM_MOUSEMOVE :
        _mx = lParam & 0xFFFF;
        _my = (lParam & 0xFFFF0000) >> 16;
        ev = new MouseMotionEvent( hWnd, _mx, _my );
        dispatch( ev );
        break;

    case WM_LBUTTONDOWN :
        _mx = lParam & 0xFFFF;
        _my = (lParam & 0xFFFF0000) >> 16;
        ev = new ButtonPressEvent( hWnd, 1, _mx, _my );
        dispatch( ev );
        _mbutton |= 1L;
        break;

    case WM_LBUTTONUP :
        _mx = lParam & 0xFFFF;
        _my = (lParam & 0xFFFF0000) >> 16;

        ev =  new ButtonReleaseEvent( hWnd, 1, _mx, _my );
        dispatch( ev );
        _mbutton &= ~1L;

        break;

    case WM_LBUTTONDBLCLK :
        _mx = lParam & 0xFFFF;
        _my = (lParam & 0xFFFF0000) >> 16;
        ev = new DoubleButtonPressEvent( hWnd, 1, _mx, _my );
        dispatch( ev );
        break;

    case WM_RBUTTONDOWN :
        _mx = lParam & 0xFFFF;
        _my = (lParam & 0xFFFF0000) >> 16;
        ev = new ButtonPressEvent( hWnd, 3, _mx, _my );
        dispatch( ev );
        _mbutton |= 4L;

        break;

    case WM_RBUTTONUP  :
        _mx = lParam & 0xFFFF;
        _my = (lParam & 0xFFFF0000) >> 16;
        ev = new ButtonReleaseEvent( hWnd, 3, _mx, _my );
        dispatch( ev );
        _mbutton &= ~4L;

        break;

    case WM_RBUTTONDBLCLK : 
        _mx = lParam & 0xFFFF;
        _my = (lParam & 0xFFFF0000) >> 16;
        dispatch( new DoubleButtonPressEvent( hWnd, 3, _mx, _my ));
        break;

    case WM_MBUTTONDOWN  :
        _mx = lParam & 0xFFFF;
        _my = (lParam & 0xFFFF0000) >> 16;
        ev = new ButtonPressEvent( hWnd, 2, _mx, _my );
        dispatch( ev );
        _mbutton |= 2L;

        break;

    case WM_MBUTTONUP :
        _mx = lParam & 0xFFFF;
        _my = (lParam & 0xFFFF0000) >> 16;
        ev = new ButtonReleaseEvent( hWnd, 2, _mx, _my );
        dispatch( ev );
        _mbutton &= ~2L;

        break;

    case WM_MBUTTONDBLCLK :
        _mx = lParam & 0xFFFF;
        _my = (lParam & 0xFFFF0000) >> 16;
        ev = new DoubleButtonPressEvent( hWnd, 2, _mx, _my );
        dispatch( ev );
        break;

    case WM_PAINT:
    {
        PAINTSTRUCT    ps;
        BeginPaint(hWnd, &ps);
        EndPaint(hWnd, &ps);
        break;
    }

    case WM_CREATE:
    case WM_SIZE:
        GetClientRect(hWnd, &rect);
        if( !_isFullScreen )
        {
            _windowX = rect.left;
            _windowY = rect.bottom;
            _windowWidth = rect.right - rect.left;
            _windowHeight = rect.bottom - rect.top;

            ev = new WindowConfigEvent( hWnd, _windowX, _windowY,
                                        _windowWidth, _windowHeight );
        }
        else
        {
            ev = new WindowConfigEvent( hWnd, rect.left, rect.bottom,
                                        rect.right - rect.left,
                                        rect.bottom - rect.top);

        }
        dispatch( ev );
        break;

    case WM_CLOSE:
        DestroyWindow (hWnd);		
        break;

    case WM_DESTROY:
        //KillGLWindow();
        PostQuitMessage (0);
        _realized = false;
		ev = new ShutdownEvent(hWnd);
		dispatch(ev);
		break;

    case WM_SYSKEYDOWN:
    case WM_KEYDOWN:
        {
            KeyCharacter keychar = Keyboard::getSingleton()->win32KeyEventToKeyChar( lParam, wParam );
            ev = new KeyPressEvent(hWnd, keychar);
            dispatch( ev );
        }
        break;

    case WM_SYSKEYUP:
    case WM_KEYUP:
        {
            KeyCharacter keychar = Keyboard::getSingleton()->win32KeyEventToKeyChar( lParam, wParam );
            ev = new KeyReleaseEvent(hWnd, keychar);
            dispatch( ev );
        }
        break;
        /*
    case WM_CHAR:
        if (lParam >> 31) {     // this bit will be set if this event is a key release
            ev = new KeyReleaseEvent(wParam);
        } else {
            ev = new KeyPressEvent(wParam);
        }
        dispatch( ev );

        if( _use_default_esc && wParam == VK_ESCAPE )
        {
            PostQuitMessage(0);
            return 0;
        }
        */

    case WM_TIMER:
        ev = new TimerEvent(hWnd, wParam);
        dispatch( ev );
        break;

    case WM_SETCURSOR:
    {
        long lRet = TRUE;
        if (LOWORD(lParam) == HTCLIENT)
        {
            switch( _curCursor )
            {
                MAP_CURSOR( GLUT_CURSOR_RIGHT_ARROW, IDC_ARROW     );
                MAP_CURSOR( GLUT_CURSOR_LEFT_ARROW,  IDC_ARROW     );
                MAP_CURSOR( GLUT_CURSOR_INFO,        IDC_HELP      );
                MAP_CURSOR( GLUT_CURSOR_DESTROY,     IDC_CROSS     );
                MAP_CURSOR( GLUT_CURSOR_HELP,        IDC_HELP       );
                MAP_CURSOR( GLUT_CURSOR_CYCLE,       IDC_SIZEALL   );
                MAP_CURSOR( GLUT_CURSOR_SPRAY,       IDC_CROSS     );
                MAP_CURSOR( GLUT_CURSOR_WAIT,           IDC_WAIT      );
                MAP_CURSOR( GLUT_CURSOR_TEXT,        IDC_UPARROW   );
                MAP_CURSOR( GLUT_CURSOR_CROSSHAIR,   IDC_CROSS     );
                ZAP_CURSOR( GLUT_CURSOR_NONE,        NULL       );

                default:
                    MAP_CURSOR( GLUT_CURSOR_UP_DOWN,     IDC_ARROW     );
            }
            break;
        }
        else
        {
            lRet = DefWindowProc( hWnd, uMsg, wParam, lParam );
        }
        break;
    }

    default:
        lRet = DefWindowProc (hWnd, uMsg, wParam, lParam);
        break;
    }
    return lRet;
} 

bool RenderSurface::realize( VisualChooser *vc, GLContext shared_context )
{
    shared_context = shared_context;
    if( _realized ) return _realized;
    if( vc != NULL )
        _visualChooser = vc;
    if( _useConfigEventThread )
    {
        _threadReady = new Barrier(2);
        startThread();
        _threadReady->block();
        delete _threadReady;
        _threadReady = 0;
    }
        
    makeCurrent();

    _realized = true;
    _realizeBlock.release();

    if( _realizeCB.valid() )
        (*_realizeCB.get())( *this );
    return (_realized = true);
}


static void doInternalError( char *msg )
{

    LPVOID lpMsgBuf;
    FormatMessage(   FORMAT_MESSAGE_ALLOCATE_BUFFER |
                     FORMAT_MESSAGE_FROM_SYSTEM |
                     FORMAT_MESSAGE_IGNORE_INSERTS,
                     NULL,
                     GetLastError(),
                     MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
                     (LPTSTR) &lpMsgBuf,\
                     0,NULL 
                     );
    char *szMessage = new char[strlen(msg)+5+strlen("Producer: Internal Error \n ")+2];
    sprintf( szMessage, "Producer: Internal Error \n %s", msg );
    MessageBox( NULL, (LPCTSTR)lpMsgBuf, szMessage, MB_OK | MB_ICONINFORMATION );
    LocalFree( lpMsgBuf );
    delete szMessage;
}


void RenderSurface::KillGLWindow()
{
    if (_glcontext)
    {
        makeCurrent();
    }

    if (_isFullScreen)
    {
        // Switch Back To The Desktop
        ChangeDisplaySettings(NULL,0);    
        ShowCursor(TRUE);
    }

    if (_glcontext && !wglDeleteContext(_glcontext))
    {
        doInternalError("Release Rendering Context Failed.");
    }
    _glcontext = NULL;

    if (_hdc && !ReleaseDC(_win,_hdc))
    {
        doInternalError("Release Device Context Failed.");
    }
    _hdc = NULL;

    if (_ownWindow)
    {
        if (_win && !DestroyWindow(_win))
        {
            doInternalError("Could Not Release hWnd.");
        }
        _win = NULL;
    }
    
    if (!UnregisterClass(_windowName.c_str(),_hinstance))
    {
        doInternalError("Could Not Unregister Class.");
    }
    _hinstance = NULL;
}

bool RenderSurface::_init()
{
    _hinstance    = GetModuleHandle(NULL);
    
    RECT rect;
    DEVMODE dm;
    memset(&dm,0,sizeof(dm));
    dm.dmSize = sizeof(dm);
    EnumDisplaySettings(NULL, ENUM_CURRENT_SETTINGS, &dm);

    _screenWidth  = dm.dmPelsWidth;
    _screenHeight = dm.dmPelsHeight;

    _setCursor(GLUT_CURSOR_LEFT_ARROW);

    if (_ownWindow )
    {
        DWORD dwExStyle;
        DWORD dwStyle;
        WNDCLASS wndclass;
        
        wndclass.style         = CS_HREDRAW | CS_VREDRAW | CS_OWNDC; 
        wndclass.lpfnWndProc   = (WNDPROC)s_proc;
        wndclass.cbClsExtra    = 0;
        wndclass.cbWndExtra    = 0;
        wndclass.hInstance     = _hinstance;
        wndclass.hCursor       = LoadCursor (NULL,IDC_ARROW);
        wndclass.hIcon         = LoadIcon( 0, "RegisterClass will fail without this non-existent ICON" );
        wndclass.hbrBackground = (HBRUSH)(COLOR_WINDOW+1);
        wndclass.lpszMenuName  = 0;
    
        wndclass.lpszClassName = _windowName.c_str();

        RegisterClass(&wndclass);
    
        GetWindowRect( GetDesktopWindow(), &rect );

        if( _windowWidth == 0xFFFFFFFF ) {
            _desiredWidth = _windowWidth = (rect.right - rect.left)/2;
            _desiredX = _windowX = _desiredWidth/2;
        } else {
            _desiredWidth = _windowWidth;
            _desiredX = _windowX;
        }

        if( _windowHeight == 0xFFFFFFFF ) {
            _desiredHeight = _windowHeight = (rect.bottom - rect.top)/2;
            _desiredY = _windowY = _desiredHeight/2;
        } else {
            _desiredHeight = _windowHeight;
            _desiredY = _windowY;
        }

        if (_isFullScreen)
        {
            if (ChangeDisplaySettings(&dm,CDS_FULLSCREEN) != DISP_CHANGE_SUCCESSFUL)
            {
                // If The Mode Fails, Offer Two Options.  Quit Or Use Windowed Mode.
                if (MessageBox(NULL,"The Requested Fullscreen Mode Is Not Supported By\nYour Video Card. Use Windowed Mode Instead?","",MB_YESNO|MB_ICONEXCLAMATION)==IDYES)
                {
                    _isFullScreen = FALSE;
                }
                else
                {
                    doInternalError("Program Will Now Close.");
                    return FALSE;
                }
            }
        }
        if (_isFullScreen)
        {
            dwExStyle = WS_EX_APPWINDOW;
            dwStyle   = WS_POPUP | WS_MAXIMIZE;

            /* windowed mode sizes remembered in _desired_$VARS */
            _windowX = 0;
            _windowY = 0;
            _windowWidth  = _screenWidth;
            _windowHeight = _screenHeight;
        }
        else
        {
            dwExStyle = WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;
            dwStyle   = WS_OVERLAPPEDWINDOW;
        }
        
        rect.left   = (long)_windowX;
        rect.right  = (long)_windowWidth;   
        
        rect.top    = (long)_windowY;
        rect.bottom = (long)_windowHeight;
        AdjustWindowRectEx(&rect, dwStyle, FALSE, dwExStyle);

        if (!(_win = CreateWindowEx(dwExStyle,
                                    _windowName.c_str(),
                                    TEXT( _windowName.c_str() ),
                                    dwStyle | WS_CLIPSIBLINGS | WS_CLIPCHILDREN,
                                    _windowX,
                                    _windowY,
                                    _windowWidth,
                                    _windowHeight,
                                    NULL,
                                    NULL,
                                    _hinstance,
                                    NULL)))
        {
            doInternalError("Window Creation Error.");
            KillGLWindow();
            return FALSE;
        }
    }
    else if( (_windowWidth == 0xFFFFFFFF) && (_windowHeight == 0xFFFFFFFF) )
    {
        GetClientRect(_win, &rect);
        _desiredX = _windowX = rect.left;
        _desiredY = _windowY = rect.bottom;
        _desiredWidth = _windowWidth = rect.right - rect.left;
        _desiredHeight = _windowHeight = rect.bottom - rect.top;
        SetWindowLong(_win, GWL_WNDPROC, (long)((WNDPROC)s_proc));
    }

    if (!(_hdc=GetDC(_win)))
    {
        doInternalError("Can't Create A GL Device Context.");
        KillGLWindow();
        return FALSE;
    }

    unsigned int visualId;
    if (_visualChooser == NULL )
    {
        VisualChooser lvc;
        lvc.setSimpleConfiguration();
        lvc.choose(&_hdc, _screen);
        visualId = lvc.getVisualID();
    }
    else
    {
        _visualChooser->choose(&_hdc,_screen);
        visualId = _visualChooser->getVisualID();
    }

    PIXELFORMATDESCRIPTOR pfd;
    if (!SetPixelFormat(_hdc, visualId, &pfd))
    {
        doInternalError("Can't Set The PixelFormat.");
        KillGLWindow();
        return FALSE;
    }

    if (!(_glcontext = wglCreateContext( _hdc )))
    {
        doInternalError("Can't Create A GL Rendering Context.");
        KillGLWindow();
        return FALSE;
    }

// ?? WHY DOES THIS CRASH HERE ??  nhv
//    if (!makeCurrent())
//    {
//        KillGLWindow();
//        MessageBox(NULL,"Can't Activate The GL Rendering Context.","ERROR",MB_OK|MB_ICONEXCLAMATION);
//        return FALSE;
//    }


    _setCursor();
    ShowWindow(_win, SW_NORMAL );
    UpdateWindow(_win );
    SetForegroundWindow(_win);
    SetFocus(_win);

    registry.insert(std::pair<Window,RenderSurface *>(_win, this));

    return true;
}


bool RenderSurface::makeCurrent( void )

{
    return(wglMakeCurrent(_hdc, _glcontext));
}

void RenderSurface::fullScreen( bool flag )
{
    if( flag == _isFullScreen )
        return;

    _isFullScreen = flag;

    if (_isFullScreen)
    {
        RECT r;
        POINT p;

        /* This hack causes the window to go back to the right position
        when it is taken out of fullscreen mode. */
        p.x = 0;
        p.y = 0;
        ClientToScreen(_win, &p);
        _desiredX = p.x;
        _desiredY = p.y;
        
        GetClientRect(_win, &r);
        _desiredWidth = r.right-r.left;
        _desiredHeight = r.bottom-r.top;

        /* force fullscreen mode */
        GetWindowRect(GetDesktopWindow(), &r);
        DWORD s = GetWindowLong(_win, GWL_STYLE);
        s &= ~WS_OVERLAPPEDWINDOW;
        s |= WS_POPUP;
        SetWindowLong(_win, GWL_STYLE, s);
        SetWindowPos(_win, 
                     HWND_TOP,
                     r.left, r.top, 
                     r.right-r.left, r.bottom-r.top, 
                     SWP_FRAMECHANGED);

    } else {
        UINT flags = SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOOWNERZORDER | SWP_NOSENDCHANGING |
                     SWP_NOSIZE | SWP_NOZORDER;

        RECT changes;
        GetClientRect(_win, &changes);
        DWORD style = GetWindowLong(_win, GWL_STYLE);

        /* Get rid of fullscreen mode, if it exists */
        if ( style & WS_POPUP ) {
            style &= ~WS_POPUP;
            style |= WS_OVERLAPPEDWINDOW;
            SetWindowLong(_win, GWL_STYLE, style);
            flags |= SWP_FRAMECHANGED;
        }

        changes.left = _desiredX;
        changes.top = _desiredY;
        flags &= ~SWP_NOMOVE;

        changes.right = changes.left + _desiredWidth;
        changes.bottom = changes.top + _desiredHeight;
        flags &= ~SWP_NOSIZE;
            /* XXX If overlay exists, resize the overlay here, ie.
        if (window->overlay) ... */

        AdjustWindowRect(&changes, style, FALSE);

        SetWindowPos(_win,
                     HWND_TOP,
                     changes.left,
                     changes.top,
                     changes.right - changes.left,
                     changes.bottom - changes.top,
                     flags);
    }
}

bool RenderSurface::makeCurrent( void ) const
{
    return(wglMakeCurrent( _hdc, _glcontext ));
}

void RenderSurface::swapBuffers( void )
{
    SwapBuffers( _hdc );
}

EventQueue * RenderSurface::selectInput( unsigned int mask )
{
    Client *client = new Client(mask);
    clients.push_back( client );
    // Need to inform clients that the windows has already
    // been created and what its size is
    if( _realized )
    {        client->queue( new WindowConfigEvent( _win,
            _windowX,
            _windowY,
            _windowWidth,
            _windowHeight ));
    }
    return client->getQueue();
}

void RenderSurface::sync( int divisor )
{
    divisor = divisor;
    glFinish();
}


void *RenderSurface::run( void )
{
    _init();
    if( _threadReady != NULL )
        _threadReady->block();

    MSG msg;

    while( GetMessage( &msg, NULL, 0, 0))
    {
        TranslateMessage( &msg );
        DispatchMessage( &msg );
    }

    return 0L;
}

void RenderSurface::_resizeWindow()
{
    RECT WindowRect;
    WindowRect.left   = long(_windowX);
    WindowRect.right  = long((_windowX + _windowWidth));
    WindowRect.top    = long((_screenHeight - (_windowY + _windowHeight)));
    WindowRect.bottom = long((_screenHeight - _windowY));
    
    AdjustWindowRect( &WindowRect,  0, false );

    SetWindowPos(_win,HWND_TOP, _windowX, WindowRect.top, WindowRect.right - WindowRect.left,
                 WindowRect.bottom - WindowRect.top, SWP_SHOWWINDOW | SWP_FRAMECHANGED);
}

void RenderSurface::_setBorder( bool flag )
{
    std::cout << "RenderSurface::_setBorder( "<< flag << " ) : Warning - Not implemented yet\n";

}

void RenderSurface::_setCursor()
{
    static int prevCursor;
    static int been_here = 0;
    
    if (!been_here) {
        prevCursor = _curCursor;
        been_here = 1;
    }
    
    POINT point;
    GetCursorPos(&point);
    if (_win == WindowFromPoint(point))
    {
        if( _useCursor )
        {
            _setCursor(prevCursor);
        } else {
            prevCursor = _curCursor;
            _setCursor(GLUT_CURSOR_NONE);
        }
    }
}

void RenderSurface::_setCursor(int cursorID)
{
    /*
     * This is a temporary solution only...
     */
    switch( cursorID )
    {
        MAP_CURSOR( GLUT_CURSOR_RIGHT_ARROW, IDC_ARROW     );
        MAP_CURSOR( GLUT_CURSOR_LEFT_ARROW,  IDC_ARROW     );
        MAP_CURSOR( GLUT_CURSOR_INFO,        IDC_HELP      );
        MAP_CURSOR( GLUT_CURSOR_DESTROY,     IDC_CROSS     );
        MAP_CURSOR( GLUT_CURSOR_HELP,        IDC_HELP       );
        MAP_CURSOR( GLUT_CURSOR_CYCLE,       IDC_SIZEALL   );
        MAP_CURSOR( GLUT_CURSOR_SPRAY,       IDC_CROSS     );
        MAP_CURSOR( GLUT_CURSOR_WAIT,         IDC_WAIT      );
        MAP_CURSOR( GLUT_CURSOR_TEXT,        IDC_UPARROW   );
        MAP_CURSOR( GLUT_CURSOR_CROSSHAIR,   IDC_CROSS     );
        ZAP_CURSOR( GLUT_CURSOR_NONE,        NULL       );

        default:
            MAP_CURSOR( GLUT_CURSOR_UP_DOWN,     IDC_ARROW     );
    }
    /*
     * Remember the currently selected cursor
     */
    _curCursor = cursorID;
}

void RenderSurface::_positionPointer(int x, int y)
{
    POINT coords = { x, y };

    // First find the new screen-relative coordinates of the mouse
    ClientToScreen( _win, &coords );

    // Now set the new mouse cursor position...
    SetCursorPos( coords.x, coords.y );
}

   
#endif  // WIN32
